/*
 *      NOMAJORS...Expected percentage of acceptable lots in single,
 *      double, or triple stage TRINOMIAL sampling in which acceptance
 *      implies that NO MAJOR nonconformances were observed.
 *
 *      Values of percentages of MAJOR and MINOR nonconforming units
 *      incoming for inspection are to be
 *
 *          ...entered from the keyboard console in response to prompts, or
 *
 *          ...redirected on the command line to come from an ASCII file.
 */

#include <stdio.h>

#define NCONFVAL 99      /* max number of incoming % nonconforming values */

int            stages, samsiz1, accnum1, reject1;
int            samsiz2, accnum2, reject2, samsiz3, accnum3;
double         pa, pb, q, lp, lq;
double         testeff, majnc, minnc, ncalast, ncblast;
double         accprob[NCONFVAL], majornc[NCONFVAL], minornc[NCONFVAL];
FILE           *outfile;

/* Logarithm of the Combinatorial Coefficient */

/*
 * NOTE: The Combinatorial Coefficient for choosing k objects from
 *       n is identical to the Trinomial Coefficient for Choosing ZERO
 *       and k objects from n...i.e. there's only 1 way to pick NONE.
 */

double logcomb( n, k )
int n, k; {

        int i;
        double  dk, dn, index, logc, log();

        logc = 0;
        if( k >= n || k <= 0 ) return( logc );

        if( n - k < k ) k = n - k;

        dk = (double)k;
        dn = (double)n;

        for( i=0; i<k; i++ ) {
                index = (double)i;
                logc += log( dn - index ) - log( dk - index );
                }

        return( logc );

        }  /* end logcomb */


/* Acceptance Probability Calculation for Single Stage Sampling */

double cumprob( accnum, samsiz )
int accnum, samsiz; {
        int    index;
        double sprob, cprob;
        double exp(), logcomb();

        /*
         * CUMPROB will be set equal to the probability that a TRINOMIAL
         * random variable with Major Nonconformance probability pa and
         * Minor Nonconformance probability pb in each of samsiz
         * independent trials will yield Zero Major and at most accnum Minor
         * nonconformances, where -
         * pa, pb, q=1-pa-pb, lp=log(pb), & lq=log(q) are in common storage.
         */

        cprob = 0;

        if( accnum < 0 || samsiz < 0 || accnum > samsiz ) return( cprob );

        for( index=0; index<=accnum; index++) {

                sprob = exp( logcomb( samsiz, index ) +
                             (double)index * lp +
                             (double)(samsiz-index) * lq );

                if( sprob > 1 )
                        sprob = 1;
                else if( sprob < 0 )
                        sprob = 0;

                cprob += sprob;
                }

        if( cprob > 1 )
               cprob = 1;
        else if( cprob < 0 )
               cprob = 0;

        /* printf( "\nCumprob for %6d, %6d = %6.4lf", accnum, samsiz, cprob); */

        return( cprob );

        }  /* end cumprob */


double sglprob( accnum, samsiz )
int accnum, samsiz; {
        double sprob;
        double exp(), logcomb();

        /*
         * SGLPROB will be set equal to the probability that a TRINOMIAL
         * random variable with Major Nonconformance probability pa and
         * Minor Nonconformance probability pb in each of samsiz
         * independent trials will yield Zero Major and exactly accnum Minor
         * nonconformances, where -
         * pa, pb, q=1-pa-pb, lp=log(pb), & lq=log(q) are in common storage.
         */
        sprob = 0;

        if( accnum < 0 || accnum > samsiz ) return( sprob );

        sprob = exp( logcomb( samsiz, accnum ) +
                     (double)accnum * lp +
                     (double)(samsiz-accnum) * lq );

        if( sprob > 1 )
                sprob = 1;
        else if( sprob < 0 )
                sprob = 0;

        /* printf( "\nSglprob for %6d, %6d = %6.4lf", accnum, samsiz, sprob); */

        return( sprob );

        }  /* end sglprob */


/* Acceptance Probability Calculation for Two Stage Sampling */

double dublsam( newacc, newsam, accept, reject, samsiz )
int newacc, newsam, accept, reject, samsiz; {
        int    index;
        double prob1, prob2, probacc, sglprob(), cumprob();

        probacc = cumprob( accept, samsiz );

        for( index=accept+1; index<reject; index++) {
                prob2 = cumprob( newacc-index, newsam );
                prob1 = sglprob( index, samsiz );
                probacc += prob1*prob2;
                }

        if( probacc > 1 )
                 probacc = 1;
        else if( probacc < 0 )
                 probacc = 0;

        /* printf( "\nDublsam for %6d, %6d, %6d, %6d, %6d = %6.4lf",
                newacc, newsam, accept, reject, samsiz, probacc); */

        return( probacc );

        }  /* end dublsam */


/* Acceptance Probability Calculation for Three Stage Sampling */

double tripsam() {
        int index;
        double prob1, prob23, probacc, sglprob(), dublsam();

        probacc = cumprob( accnum1, samsiz1 );

        for( index=accnum1+1; index<reject1; index++) {
                prob23   = dublsam( accnum3-index, samsiz3,
                                    accnum2-index, reject2, samsiz2 );
                prob1    = sglprob( index, samsiz1 );
                probacc += prob1*prob23;
                }

        if( probacc > 1 ) probacc = 1;
        else if( probacc < 0 ) probacc = 0;

        return( probacc );

        }  /* end tripsam */

main(argc, argv)
int argc;
char *argv[]; {
        char           buf[100];
        int            index, i;
        double cumprob(), dublsam(), tripsam();
        extern double log();
        extern FILE *fopen();                   /* Stream I/O */
        extern int fclose();                    /* Stream Output */
        extern FILE *outfile;
        unsigned int   c;

        printf("\n\n*** One, Two, or Three Stage TRINOMIAL Acceptance ***");
        printf("\n    Sampling in which the Observance of a Single");
        printf("\n    Major Nonconformance implies LOT Rejection.\n");
        printf("\n         A Quality Assurance Training Tool:");
        printf("\n  Statistics Committee of the QA Section of the PMA.");
        printf("\n\n  by Bob Obenchain, Manager, Nonclinical Statistics");
        printf("\n         Glaxo, MIS Statistical Services.\n\n");

        if( argc >= 2 ) {
                printf("...The output file will be: %s\n", argv[1]);
                if( ( outfile = fopen(argv[1], "w") ) == NULL ) {
                    printf("\n\tCannot open output file: %s\n", argv[1]);
                    if( ( outfile = fopen("nomajors.out", "w") ) == NULL ) {
                        printf("\tCannot open default output file, nomajors.out.\n");
                        exit( 1 );
                        }
                    else printf("\t...using default output filename, nomajors.out.\n");
                    }
                }
        else {
                if( (outfile = fopen("nomajors.out", "w") ) == NULL) {
                        printf("\tCannot open default output file, nomajors.out.\n");
                        exit(1);
                        }
                else printf("...using output filename = nomajors.out...\n");
                }

        stages = 1;

        printf("\n\nEnter First Stage Sample Size: ");
        fgets(buf, 100, stdin);
        sscanf(buf, "%d", &samsiz1);
        if( samsiz1 <= 0 ) {
                printf("\nFirst Stage Sample Size will be 50, the default.\n");
                samsiz1 = 50;
                }
        else if( samsiz1 == 1 ) {
                printf("\nFirst Stage Sample Size will be 2, the minimum allowed.\n");
                samsiz1 = 2;
                }
        else if( samsiz1 > 5000 ) {
                printf("\nFirst Stage Sample Size will be 5,000...which is the maximum allowed here.\n");
                samsiz1 = 5000;
                }

        printf("\nEnter First Stage MINOR Nonconformance Accept Number: ");
        fgets(buf, 100, stdin);
        sscanf(buf, "%d", &accnum1);
        if( accnum1 < -1 ) {
                printf("\nStage 1 Accept Number will be 0, the default.\n");
                accnum1 = 0;
                }
        else if( accnum1 == -1 ) {
                printf("\nA Stage 1 Accept Number of -1 indicates that the lot cannot be accepted\n");
                printf("at this first stage even if zero nonconformances are observed.\n");
                }
        else if( accnum1 >= samsiz1 ) {
                accnum1 = samsiz1/4;
                printf("\nStage 1 Accept Number will be %d, which is about one-fourth the Sample Size of %d.\n", accnum1, samsiz1);
                }
        if( accnum1 > 999 ) {
                printf("\nAccept Numbers cannot exceed 999.\n");
                accnum1 = 999;
                }

        printf("\nEnter First Stage MINOR Nonconformance Reject Number: ");
        fgets(buf, 100, stdin);
        sscanf(buf, "%d", &reject1);
        if( reject1 < 1 ) reject1 = 1;
        if( reject1 <= accnum1 + 1 ) {
                if( accnum1 < 0 ) {
                        printf("\nAccept Number reset to Zero for Single Stage Sampling.\n");
                        accnum1 = 0;
                        }
                reject1 = accnum1 + 1;
                if( reject1 > samsiz1 ) {
                        reject1 = samsiz1;
                        printf("\nStage 1 Reject Number will be %d, ...all sampled units nonconform.\n", reject1);
                        }
                else {
                        printf("\nA Reject Number of %d is the minimum allowed for\n", reject1);
                        printf("Single Stage Sampling with Accept Number of %d.\n", accnum1);
                        }
                }

        if( reject1 > accnum1 + 1 ) {   /* Second Stage Required */

                stages = 2;

                printf("\nEnter Second Stage Sample Size: ");
                fgets(buf, 100, stdin);
                sscanf(buf, "%d", &samsiz2);
                if( samsiz2 <= 0 ) {
                        printf("\nSecond Stage Sample Size will be 50, the default.\n");
                        samsiz2 = 50;
                        }
                else if( samsiz2 > 5000 ) {
                        printf("\nSecond Stage Sample Size will be 5,000...which is the maximum allowed here.\n");
                        samsiz2 = 5000;
                        }

                printf("\nEnter Second Stage MINOR Nonconformance Accept Number: ");
                fgets(buf, 100, stdin);
                sscanf(buf, "%d", &accnum2);
                if( accnum2 < accnum1 ) {
                        accnum2 = accnum1;
                        printf("\nSecond Stage Accept Number will be %d, which is the minimum allowed when\n", accnum2);
                        printf("the First Stage Accept Number is %d.\n", accnum1);
                        }
                if( accnum2 < reject1 - 1 ) {
                        printf("\nA Third Stage is required when the Second Stage Accept Number is");
                        printf("\ntwo or more units smaller than the First Stage Reject Number.");
                        }
                if( accnum2 >= samsiz1 + samsiz2 ) {
                        accnum2 = (samsiz1 + samsiz2)/4;
                        printf("\nSecond Stage Accept Number will be %d, which is about one-fourth of the", accnum2);
                        printf("\n1st and 2nd Stage Sample Size of %d.\n", samsiz1+samsiz2);
                        }
                if( accnum2 > 999 ) {
                        printf("\nAccept Numbers cannot exceed 999.\n");
                        accnum2 = 999;
                        }

                printf("\nEnter Second Stage MINOR Nonconformance Reject Number: ");
                fgets(buf, 100, stdin);
                sscanf(buf, "%d", &reject2);
                if( reject2 <= accnum2 + 1 ) {
                        reject2 = accnum2 + 1;
                        if( reject2 > samsiz1 + samsiz2 ) {
                                reject1 = samsiz1 + samsiz2;
                                printf("\nStage 2 Reject Number will be %d, ...all sampled units nonconform.\n", reject2);
                                }
                        else {
                                printf("\nA Reject Number of %d is the minimum allowed for\n", reject2);
                                printf("Two Stage Sampling with Accept Number of %d.\n", accnum2);
                                }
                        }
                }

        if( reject2 > accnum2 + 1 ) {   /* Third Stage Required */

                stages = 3;

                printf("\nEnter Third Stage Sample Size: ");
                fgets(buf, 100, stdin);
                sscanf(buf, "%d", &samsiz3);
                if( samsiz3 <= 0 ) {
                        printf("\nThird Stage Sample Size will be 50, the default.\n");
                        samsiz3 = 50;
                        }
                else if( samsiz3 > 5000 ) {
                        printf("\nThird Stage Sample Size will be 5,000...which is the maximum allowed here.\n");
                        samsiz3 = 5000;
                        }

                printf("\nEnter Third Stage MINOR Nonconformance Accept Number: ");
                fgets(buf, 100, stdin);
                sscanf(buf, "%d", &accnum3);
                if( accnum3 < reject2 - 1 ) {
                        accnum3 = reject2 - 1;
                        printf("\nThe Third Stage Accept Number will be %d, which is the minimum allowed when\n", accnum3);
                        printf("the Second Stage Reject Number is %d.\n", reject2);
                        }
                if( accnum3 >= samsiz1 + samsiz2 + samsiz3 ) {
                        accnum2 = (samsiz1 + samsiz2 + samsiz3)/4;
                        printf("\nThird Stage Accept Number will be %d, which is about one-fourth of the", accnum3);
                        printf("\nThree Stage Sample Total Size of %d.\n", samsiz1+samsiz2+samsiz3);
                        }
                if( accnum3 > 999 ) {
                        printf("\nAccept Numbers cannot exceed 999.\n");
                        accnum3 = 999;
                        }
                }

        printf("\nSpecify Testing Efficiency (Percentage): ");
        fgets(buf, 100, stdin);
        sscanf(buf, "%lf", &testeff);
        testeff /= 100;
        if( testeff > 1 || testeff < 0.01 ) {
                printf("\nTesting Efficiency will be 100%%, the default.\n");
                testeff = 1;
                }
        else if( testeff < 0.1 ) {
                printf("\nTesting Efficiency will be 10%%, the minimum allowed here.\n");
                testeff = 0.1;
                }

        if( stages == 1 )
                printf("\n\nSUMMARY: One Stage Sampling...\n");
        else if( stages == 2 )
                printf("\n\nSUMMARY: Two Stage Sampling...\n");
        else
                printf("\n\nSUMMARY: Three Stage Sampling...\n");

        printf("\nFirst Stage Sample Size = %d", samsiz1 );
        fprintf( outfile, "%4d = First Stage Sample Size", samsiz1 );
        printf("\nMAJOR Nonconformance Reject Number = 1 at ALL Stages.");
        fprintf( outfile, "\nMAJOR Nonconformance Reject Number = 1 at ALL Stages.");
        printf("\nFirst Stage MINOR Nonconformance Accept Number = %d", accnum1 );
        fprintf( outfile, "\n%4d = First Stage MINOR Nonconformance Accept Number", accnum1 );
        printf("\nFirst Stage MINOR Nonconformance Reject Number = %d", reject1 );
        fprintf( outfile, "\n%4d = First Stage MINOR Nonconformance Reject Number", reject1 );

        if( stages >= 2 ) {
                printf("\nSecond Stage Sample Size = %d", samsiz2 );
                fprintf( outfile,"\n%4d = Second Stage Sample Size", samsiz2 );
                printf("\nSecond Stage MINOR Nonconformance Accept Number = %d", accnum2 );
                fprintf( outfile,"\n%4d = Second Stage MINOR Nonconformance Accept Number", accnum2 );
                printf("\nSecond Stage MINOR Nonconformance Reject Number = %d", reject2 );
                fprintf( outfile,"\n%4d = Second Stage MINOR Nonconformance Reject Number", reject2 );
                }

        if( stages == 3 ) {
                printf("\nThird Stage Sample Size = %d", samsiz3 );
                fprintf( outfile,"\n%4d = Third Stage Sample Size", samsiz3 );
                printf("\nFinal MINOR Nonconformance Accept Number = %d", accnum3 );
                fprintf( outfile,"\n%4d = Final MINOR Nonconformance Accept Number", accnum3 );
                }

        printf("\nTesting Efficiency = %6.1lf%%", 100*testeff );
        fprintf( outfile, "\n%6.1lf = Testing Efficiency %%", 100*testeff );

        printf("\n\nPress a KEY to start calculations...");
        fprintf( outfile, "\nIn%%Major In%%Minor %%Yield");
        getchar();

        ncalast = ncblast = -1;

        for(index=0; index<NCONFVAL; index++) {

                printf("\n\nStep Number = %d", index);

                printf("\n\nEnter a MAJOR Nonconformance Percentage: ");
                fgets(buf, 100, stdin);
                sscanf(buf, "%lf", &majnc);

                if( majnc < 0 ) break;

                if( majnc >= 100 ) {
                        printf("\nMAJOR Nonconformance Percentage has been set to 50%%.\n");
                        majnc = 50;
                        }

                majornc[index] = majnc;

                pa = testeff*majnc/100;  /* convert % to fraction */

                printf("\n\nEnter a MINOR Nonconformance Percentage: ");
                fgets(buf, 100, stdin);
                sscanf(buf, "%lf", &minnc);

                if( minnc < 0 ) minnc = 0;

                if( minnc >= 100 ) {
                        printf("\nMINOR Nonconformance Percentage has been set to 50%%.\n");
                        minnc = 50;
                        }

                minornc[index] = minnc;

                pb = testeff*minnc/100;  /* convert % to fraction */

                if( majnc + minnc > 100 ) {
                        printf("\nMAJOR plus MINOR Nonconformance Percentages cannot exceed 100%%.\n");
                        continue;
                        }

                if( pa == 0 && pb == 0 )
                        accprob[index] = 1;

                else if( pa + pb >= 1 )
                        accprob[index] = 0;

                else if( majnc != ncalast || minnc != ncblast ) {

                        /* perform calculations only when the incoming
                           nonconformance percentages have changed from
                           the last step. */

                        if( pb < 0.0000000001 ) pb = 0.0000000001;

                        /* to handle the cases where pa > 0 but pb zero. */

                        lp = log( pb );
                        q = 1 - pa - pb;

                        if( q <= 0 ) break; /* Should Never Happen! */

                        lq = log( q );

                        if( stages == 1 )
                                accprob[index] = cumprob( accnum1, samsiz1 );

                        else if( stages == 2 )
                                accprob[index] = dublsam( accnum2, samsiz2,
                                                 accnum1, reject1, samsiz1 );
                        else if( stages == 3 )
                                accprob[index] = tripsam();
                        }

                else accprob[index] = accprob[index-1];

                printf("\n\n For Incoming MAJOR Nonconformance Percentage = %6.2lf,",
                        majnc);
                printf("\n and Incoming MINOR Nonconformance Percentage = %6.2lf,",
                        minnc);
                printf("\n Expected First-Pass Yield Percentage   = %6.2lf,",
                        100*accprob[index]);

                if( majnc != ncalast || minnc != ncblast ) {

                        /* Successive duplicates not writtem to output file. */

                        fprintf( outfile, "\n %6.2lf %6.2lf %6.2lf",
                                 majnc, minnc, 100*accprob[index]);
                        }

                ncalast = majnc;
                ncblast = minnc;

                printf("\n\nTo QUIT, enter a NEGATIVE MAJOR Nonconformance Percentage...");
                }

        fprintf( outfile, "\n-1 ...End of Input Signal...");
        fclose(outfile);

        exit(0);

        }

